LOG_DEBUG = 4
LOG_ERROR = 1
LOG_INFO = 3
LOG_NONE = 0
LOG_VERBOSE = 5
LOG_WARNING = 2


def dht_readinto():
    pass


def flash_erase():
    pass


def flash_read():
    pass


def flash_size():
    pass


def flash_user_start():
    pass


def flash_write():
    pass


def gpio_matrix_in():
    pass


def gpio_matrix_out():
    pass


def neopixel_write():
    pass


def osdebug():
    pass
